typedef GoUint64_ httphelper__Coins;
typedef GoUint64_ httphelper__Hours;
typedef struct {
    cipher__Address _unnamed;
} httphelper__Address;

typedef struct
{
    cipher__SHA256 _unnamed;
} httphelper__SHA256;