package main

import droplet "github.com/skycoin/skycoin/src/util/droplet"

/*

  #include <string.h>
  #include <stdlib.h>

  #include "skytypes.h"
*/
import "C"

//export SKY_droplet_FromString
func SKY_droplet_FromString(_b string, _arg1 *uint64) (____error_code uint32) {
	b := _b
	__arg1, ____return_err := droplet.FromString(b)
	____error_code = libErrorCode(____return_err)
	if ____return_err == nil {
		*_arg1 = __arg1
	}
	return
}

//export SKY_droplet_ToString
func SKY_droplet_ToString(_n uint64, _arg1 *C.GoString_) (____error_code uint32) {
	n := _n
	__arg1, ____return_err := droplet.ToString(n)
	____error_code = libErrorCode(____return_err)
	if ____return_err == nil {
		copyString(__arg1, _arg1)
	}
	return
}
