// Package kvstorage contains the implementation for a
// key-value storage of arbitrary data. `Manager` is used to
// access the storage contents. Each storage is presented by its own `KVStorageType`
// and each type has its own associated file to persist data
package kvstorage
