/**
 * Hash signed using a secret key, 65 bytes long.
 */
typedef GoUint8_ cipher__Sig[65];

/**
 * Public key, 33-bytes long.
 */
typedef GoUint8_ cipher__PubKey[33];

/**
 * Secret key, 32 bytes long.
 */
typedef GoUint8_ cipher__SecKey[32];
